#ifndef RING_EXTENSIONS_H
#define RING_EXTENSIONS_H

#include "ring-extensions/_gen/svc.h"
#include "ring-extensions/_gen/enums.h"
#include "ring-extensions/_gen/gtypes.h"
#include "ring-extensions/_gen/interfaces.h"

#endif

